import re


class Error(Exception):
    """Base class for all exceptions raised by this package"""

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and str(self) == str(other)
        )


class RPCError(Error):
    """
    Generic RPC error

    This can be some kind of miscommunication with the RPC service (e.g. unknown
    method called, invalid or missing argument, etc), which should be a
    considered a bug. But it can also be a normal error message (e.g. unknown
    torrent), that should be communicated to the user.
    """

    def translate(self, map):
        # By using r-string we don't have to escape backslashes
        r"""
        Turn this exception into another one based on regular expressions

        :param map: Mapping of regular expression strings to target exception
            instances

        Each regular expression is matched aagainst the error message of the
        instance (i.e. ``str(self)``). The corresponding target exception of the
        first matching regular expression is returned.

        >>> RPCError("foo is bad").translate({
        >>>     r"^foo": ValueError("Foo is something"),
        >>>     r"is bad$": TypeError("Something is bad"),
        >>> })
        ValueError('bad: Foo')

        If there are no matches, the instance (``self``) is returned.

        >>> RPCError("foo is bad").translate({
        >>>     r"Hello?": ValueError("Anybody out there?"),
        >>> })
        RPCError("foo is bad")

        If the matching target exception is a ``(exception_class, message)``
        tuple, any group references in ``message`` are substituted with the
        groups from the matching regular expression and
        ``exception_class(message_with_group_substitutions)`` is returned.

        >>> for e in (RPCError("foo is bad"), RPCError("bar is bad"), RPCError("bar is silly")):
        >>>     e.translate({
        >>>         r"^(\w+) is (\w+)$": (ValueError, r"\2: \1"),
        >>>     })
        ValueError('bad: foo')
        ValueError('bad: bar')
        ValueError('silly: bar')
        """
        self_msg = str(self)
        for regex, to_exc in map.items():
            match = re.search(regex, self_msg)
            if match:
                if isinstance(to_exc, tuple) and len(to_exc) == 2:
                    to_cls, to_msg = to_exc
                    # Replace group references (\1, \g<1>, \g<name>) in `to_msg`
                    return to_cls(match.expand(to_msg))
                else:
                    return to_exc
        return self


class ConnectionError(Error):
    """Failed to connect to the client, e.g. because it isn't running"""

    def __init__(self, msg):
        # python_socks.ProxyConnectionError provides ugly errors messages,
        # e.g. "Could not connect to proxy localhost:1337 [None]".
        msg = re.sub(r'\s+\[.*?\]$', '', str(msg))
        super().__init__(msg)


class TimeoutError(Error, TimeoutError):
    """
    Timeout for sending request and reading response

    Besides :class:`Error`, this is also a subclass of :class:`TimeoutError`.
    """


class AuthenticationError(Error):
    """Failed to prove identity"""


class ValueError(Error, ValueError):
    """
    Invalid value (e.g. port 65536)

    Besides :class:`Error`, this is also a subclass of :class:`ValueError`.
    """



import json

from . import _base, _errors, _utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TransmissionURL(_utils.URL):
    """Transmission RPC URL"""

    default = 'http://localhost:9091/transmission/rpc'

    @property
    def scheme(self):
        """Valid schemes: ``http``, ``https``"""
        return super().scheme

    @scheme.setter
    def scheme(self, scheme):
        if scheme and scheme.lower() not in ('http', 'https'):
            raise _errors.ValueError('Scheme must be "http" or "https"')
        else:
            _utils.URL.scheme.fset(self, scheme)


class TransmissionRPC(_base.RPCBase):
    """
    RPC client for Transmission

    URL format: ``[http[s]://][USERNAME:PASSWORD@]HOST[:PORT][/PATH]``

    Reference: https://github.com/transmission/transmission/blob/master/extras/rpc-spec.txt

    **Calling RPC methods**

    RPC methods can be called like python functions:

    >>> client.call("torrent-add", filename="path/to.torrent", paused=True)

    If you need to pass argument names that contain characters that are illegal
    for keyword arguments (e.g. "-"), provide a dictionary:

    >>> client.call(
        "torrent-add",
        {"filename": "file.torrent", "download-dir": "/some/path"},
    )

    If you provide keyword arguments and a dictionary, values from the
    dictionary overload keyword arguments.

    :raise ValueError: if any argument is invalid
    """

    name = 'transmission'
    label = 'Transmission'
    URL = TransmissionURL

    def __init__(
        self,
        url=None,
        *,
        scheme=None,
        host=None,
        port=None,
        path=None,
        username=None,
        password=None,
        timeout=None,
        proxy_url=None,
    ):
        # Set custom or default URL
        self.url = url

        # Update URL
        if scheme:
            self.url.scheme = scheme
        if host:
            self.url.host = host
        if port:
            self.url.port = port
        if path:
            self.url.path = path
        if username:
            self.url.username = username
        if password:
            self.url.password = password

        self.timeout = timeout
        self.proxy_url = proxy_url

    _auth_error_code = 401
    _csrf_error_code = 409
    _csrf_header = 'X-Transmission-Session-Id'

    async def _request(self, method, tag=None, **parameters):
        data = {'method': str(method)}
        if parameters:
            data['arguments'] = parameters
        if tag:
            try:
                data['tag'] = int(float(tag))
            except (TypeError, ValueError):
                raise _errors.ValueError(f'Tag must be a number: {tag!r}')

        try:
            data_json = json.dumps(data)
        except Exception:
            raise _errors.ValueError(f'Failed to serialize to JSON: {data}')

        response = await self._send_post_request(str(self.url), data=data_json)

        if response.status_code == self._csrf_error_code and self._csrf_header in response.headers:
            # Try again with CSRF header
            _log.debug('Setting CSRF header: %s = %s', self._csrf_header, response.headers[self._csrf_header])
            self._http_headers[self._csrf_header] = response.headers[self._csrf_header]
            response = await self._send_post_request(str(self.url), data=data_json)

        if response.status_code == self._auth_error_code:
            raise _errors.AuthenticationError('Authentication failed')

        elif response.status_code != 200:
            raise RuntimeError(f'Unexpected response: {response!r}')

        return response

    async def _connect(self):
        # There is no special login procedure, we just make a valid request to
        # see if it basically works
        await self._request('session-stats')

    async def _disconnect(self):
        # Forget CSRF header
        self._http_headers.clear()

    async def _call(self, method, args=None, **kwargs):
        if args:
            kwargs.update(args)
        response = await self._request(method, **kwargs)
        try:
            # Any error message is in "result" field in the returned JSON object
            return response.json()
        except Exception:
            raise _errors.RPCError(f'Unexpected response: {response.text}')



import asyncio
import re
import ssl
import struct
import zlib
from unittest.mock import Mock, call

import pytest
import python_socks
import rencode

from aiobtclientrpc import RPCBase, _deluge, _errors, _utils

from .common import AsyncMock, make_url_parts


@pytest.mark.parametrize(
    argnames='url, exp',
    argvalues=(
        ('',
         {'scheme': None, 'host': 'localhost', 'port': '58846', 'path': None, 'username': None, 'password': None}),
        ('myhost',
         {'scheme': None, 'host': 'myhost', 'port': '58846', 'path': None, 'username': None, 'password': None}),
        ('myhost:123',
         {'scheme': None, 'host': 'myhost', 'port': '123', 'path': None, 'username': None, 'password': None}),
        ('foo:bar@myhost',
         {'scheme': None, 'host': 'myhost', 'port': '58846', 'path': None, 'username': 'foo', 'password': 'bar'}),
        ('foo:bar@myhost:123',
         {'scheme': None, 'host': 'myhost', 'port': '123', 'path': None, 'username': 'foo', 'password': 'bar'}),
        ('http://myhost',
         _errors.ValueError("Deluge URLs don't have a scheme")),
        ('myhost/foo',
         _errors.ValueError("Deluge URLs don't have a path")),
    ),
    ids=lambda v: str(v),
)
def test_DelugeURL(url, exp):
    if isinstance(exp, Exception):
        with pytest.raises(type(exp), match=rf'^{re.escape(str(exp))}$'):
            _deluge.DelugeURL(url)
    else:
        url = _deluge.DelugeURL(url)
        assert make_url_parts(url) == exp


@pytest.mark.parametrize('url', (None, 'a:b@foo:123'))
@pytest.mark.parametrize(
    argnames='kwargs',
    argvalues=(
        {},
        {'host': 'asdf'},
        {'port': '123'},
        {'username': 'this', 'password': 'that'},
        {'timeout': 123},
        {'proxy_url': 'http://hey:ho@bar:456'},
    ),
    ids=lambda v: str(v),
)
def test_DelugeRPC_instantiation(kwargs, url):
    if url:
        kwargs['url'] = url
    rpc = _deluge.DelugeRPC(**kwargs)

    default_url = _utils.URL(_deluge.DelugeURL.default)
    exp_url = _utils.URL(_deluge.DelugeURL.default)
    if url:
        custom_url = _utils.URL(url)
    for name in ('scheme', 'host', 'port', 'path', 'username', 'password'):
        if name in kwargs:
            exp_value = kwargs[name]
        elif url:
            exp_value = getattr(custom_url, name)
        else:
            exp_value = getattr(default_url, name)
        actual_value = getattr(rpc.url, name)
        assert actual_value == exp_value
        setattr(exp_url, name, exp_value)

    assert rpc.url == exp_url
    assert rpc.timeout == kwargs.get('timeout', RPCBase.default_timeout)
    if 'proxy_url' in kwargs:
        assert rpc.proxy_url == _utils.URL(kwargs['proxy_url'])
    else:
        assert rpc.proxy_url is None


@pytest.mark.asyncio
async def test_DelugeRPC_connect(mocker):
    rpc = _deluge.DelugeRPC(url='foo:bar@localhost:123')
    mocks = Mock(
        _disconnect=AsyncMock(),
        _call=AsyncMock(),
        _DelugeRPCClient=Mock(),
        catch_connection_exceptions=AsyncMock(return_value='catch_connection_exceptions return value'),
    )
    mocker.patch.object(rpc, '_disconnect', mocks._disconnect)
    mocker.patch.object(rpc, '_call', mocks._call)
    mocker.patch('aiobtclientrpc._deluge._DelugeRPCClient', mocks._DelugeRPCClient)
    mocker.patch('aiobtclientrpc._utils.catch_connection_exceptions', mocks.catch_connection_exceptions)

    rpc._event_handlers['foo'] = Mock()
    rpc._event_handlers['bar'] = Mock()

    await rpc._connect()

    assert rpc._client is mocks._DelugeRPCClient.return_value
    assert mocks.mock_calls == [
        call._disconnect(),
        call._DelugeRPCClient(
            host=rpc.url.host,
            port=rpc.url.port,
            on_connection_lost=rpc._on_connection_lost,
            proxy_url=rpc.proxy_url,
            event_handler=rpc._emit_event,
        ),
        call._DelugeRPCClient().login(
            username=rpc.url.username,
            password=rpc.url.password,
        ),
        call.catch_connection_exceptions(rpc._client.login.return_value),
        call._call('daemon.set_event_interest', ['foo']),
        call._call('daemon.set_event_interest', ['bar']),
    ]


def test_DelugeRPC_on_connection_lost():
    rpc = _deluge.DelugeRPC(url='foo:bar@localhost:123')
    disconnected_cb = Mock()
    rpc.set_disconnected_callback(disconnected_cb, 1, two=3)
    rpc._status = _utils.ConnectionStatus.connecting
    assert rpc.status is _utils.ConnectionStatus.connecting
    rpc._on_connection_lost()
    assert rpc.status is _utils.ConnectionStatus.disconnected
    assert disconnected_cb.call_args_list == [call(1, two=3)]


@pytest.mark.parametrize(
    argnames='has_client',
    argvalues=(False, True),
)
@pytest.mark.asyncio
async def test_DelugeRPC_disconnect(has_client, mocker):
    rpc = _deluge.DelugeRPC()
    catch_connection_exceptions_mock = mocker.patch('aiobtclientrpc._utils.catch_connection_exceptions',
                                                    AsyncMock(return_value='catch_connection_exceptions return value'))
    if has_client:
        rpc._client = client_mock = Mock(logout=Mock(return_value='client.logout coroutine'))

        await rpc._disconnect()

        assert catch_connection_exceptions_mock.call_args_list == [call('client.logout coroutine')]
        assert client_mock.logout.call_args_list == [call()]

    assert not hasattr(rpc, '_client')


@pytest.mark.asyncio
async def test_DelugeRPC_call(mocker):
    rpc = _deluge.DelugeRPC()
    catch_connection_exceptions_mock = mocker.patch('aiobtclientrpc._utils.catch_connection_exceptions',
                                                    AsyncMock(return_value='catch_connection_exceptions return value'))
    rpc._client = Mock(call=Mock(return_value='client.call coroutine'))

    return_value = await rpc._call('method_name', 'foo', bar='baz')

    assert return_value is catch_connection_exceptions_mock.return_value
    assert catch_connection_exceptions_mock.call_args_list == [call('client.call coroutine')]
    assert rpc._client.call.call_args_list == [call('method_name', 'foo', bar='baz')]


@pytest.mark.asyncio
async def test_DelugeRPC_subscribe(mocker):
    rpc = _deluge.DelugeRPC()
    mocker.patch.object(rpc, 'call', AsyncMock())
    await rpc._subscribe('foo_event')
    assert rpc.call.call_args_list == [call('daemon.set_event_interest', ['foo_event'])]


@pytest.mark.asyncio
async def test_DelugeRPC_unsubscribe(mocker):
    client = _deluge.DelugeRPC()
    mocker.patch.object(client, 'call', AsyncMock())
    await client._unsubscribe('foo_event')
    assert client.call.call_args_list == []


async def test_DelugeRPCClient_connection_lost():
    client = _deluge._DelugeRPCClient(host='localhost', port=123, on_connection_lost=Mock())
    client._protocol = 'foo'
    client._connection_lost()
    assert client._on_connection_lost.call_args_list == [call()]
    assert client._protocol is None


@pytest.mark.parametrize('iscoroutinefunction', (True, False))
@pytest.mark.asyncio
async def test_DelugeRPCClient_event_received(iscoroutinefunction, mocker):
    client = _deluge._DelugeRPCClient(host='localhost', port=123, event_handler=Mock())
    mocker.patch.object(client, '_loop')
    mocker.patch('asyncio.iscoroutinefunction', return_value=iscoroutinefunction)

    client._event_received('foo', (1, 2, 3))

    if iscoroutinefunction:
        assert client._event_handler.call_args_list == [call('foo', (1, 2, 3))]
        assert client._loop.create_task.call_args_list == [call(client._event_handler.return_value)]
        assert client._loop.call_soon.call_args_list == []
    else:
        assert client._loop.call_soon.call_args_list == [call(client._event_handler, 'foo', (1, 2, 3))]
        assert client._event_handler.call_args_list == []
        assert client._loop.create_task.call_args_list == []


@pytest.mark.asyncio
async def test_DelugeRPCClient_protocol_factory(mocker):
    DelugeRPCProtocol_mock = mocker.patch('aiobtclientrpc._deluge._DelugeRPCProtocol')
    client = _deluge._DelugeRPCClient(host='localhost', port=123)
    protocol = client._protocol_factory()
    assert protocol is DelugeRPCProtocol_mock.return_value
    assert DelugeRPCProtocol_mock.call_args_list == [call(
        on_connection_lost=client._connection_lost,
        on_event_received=client._event_received,
    )]


@pytest.mark.asyncio
async def test_DelugeRPCClient_create_ssl_context():
    client = _deluge._DelugeRPCClient(host='localhost', port=123)
    ctx = client._create_ssl_context()
    assert isinstance(ctx, ssl.SSLContext)
    assert ctx.check_hostname is False
    assert ctx.verify_mode is ssl.CERT_NONE


@pytest.mark.asyncio
async def test_DelugeRPCClient_login_when_already_logged_in(mocker):
    client = _deluge._DelugeRPCClient(host='localhost', port=123)
    client._protocol = 'mock protocol'

    Proxy_from_url_mock = mocker.patch('python_socks.async_.asyncio.Proxy.from_url', return_value=Mock(
        connect=AsyncMock(),
    ))
    mocker.patch.object(client, '_create_ssl_context')
    catch_connection_exceptions_mock = mocker.patch('aiobtclientrpc._utils.catch_connection_exceptions',
                                                    AsyncMock(return_value=('mock transport', 'mock protocol')))
    mocker.patch.object(client._loop, 'create_connection', Mock())
    mocker.patch.object(client, 'call', AsyncMock())

    await client.login('Username', 'Password')

    assert Proxy_from_url_mock.call_args_list == []
    assert Proxy_from_url_mock.return_value.connect.call_args_list == []
    assert catch_connection_exceptions_mock.call_args_list == []
    assert client._loop.create_connection.call_args_list == []
    assert client.call.call_args_list == []

@pytest.mark.asyncio
async def test_DelugeRPCClient_login_with_proxy_url(mocker):
    client = _deluge._DelugeRPCClient(
        host='localhost',
        port=123,
        proxy_url=_utils.URL('socks5://localhost:456'),
    )
    client._protocol = None

    Proxy_from_url_mock = mocker.patch('python_socks.async_.asyncio.Proxy.from_url', return_value=Mock(
        connect=AsyncMock(),
    ))
    mocker.patch.object(client, '_create_ssl_context')
    catch_connection_exceptions_mock = mocker.patch('aiobtclientrpc._utils.catch_connection_exceptions',
                                                    AsyncMock(return_value=('mock transport', 'mock protocol')))
    mocker.patch.object(client._loop, 'create_connection', Mock())
    mocker.patch.object(client, 'call', AsyncMock(side_effect=(
        'mock client_version',
        'mock login response',
    )))

    await client.login('Username', 'Password')

    assert Proxy_from_url_mock.call_args_list == [call(client._proxy_url.with_auth)]
    assert Proxy_from_url_mock.return_value.connect.call_args_list == [call(
        dest_host=client._host,
        dest_port=client._port,
        timeout=float('inf'),
    )]
    assert catch_connection_exceptions_mock.call_args_list == [call(
        client._loop.create_connection.return_value,
    )]
    assert client._loop.create_connection.call_args_list == [call(
        sock=Proxy_from_url_mock.return_value.connect.return_value,
        server_hostname=client._host,
        protocol_factory=client._protocol_factory,
        ssl=client._create_ssl_context.return_value,
        ssl_handshake_timeout=float('inf'),
    )]
    assert client.call.call_args_list == [
        call('daemon.info'),
        call('daemon.login', 'Username', 'Password', client_version='mock client_version'),
    ]

@pytest.mark.asyncio
async def test_DelugeRPCClient_login_with_bad_proxy_url(mocker):
    client = _deluge._DelugeRPCClient(
        host='localhost',
        port=123,
        proxy_url=_utils.URL('socks5://localhost:456'),
    )
    client._protocol = None

    Proxy_from_url_mock = mocker.patch('python_socks.async_.asyncio.Proxy.from_url',
                                       side_effect=ValueError('bad proxy url'))
    mocker.patch.object(client, '_create_ssl_context')
    catch_connection_exceptions_mock = mocker.patch('aiobtclientrpc._utils.catch_connection_exceptions',
                                                    AsyncMock(return_value=('mock transport', 'mock protocol')))
    mocker.patch.object(client._loop, 'create_connection', Mock())
    mocker.patch.object(client, 'call', AsyncMock(side_effect=(
        'mock client_version',
        'mock login response',
    )))

    with pytest.raises(_errors.ValueError, match=r'^bad proxy url$'):
        await client.login('Username', 'Password')

    assert Proxy_from_url_mock.call_args_list == [call(client._proxy_url.with_auth)]
    assert Proxy_from_url_mock.return_value.connect.call_args_list == []
    assert catch_connection_exceptions_mock.call_args_list == []
    assert client._loop.create_connection.call_args_list == []
    assert client.call.call_args_list == []

@pytest.mark.asyncio
async def test_DelugeRPCClient_login_with_bad_proxy_connection(mocker):
    client = _deluge._DelugeRPCClient(
        host='localhost',
        port=123,
        proxy_url=_utils.URL('socks5://localhost:456'),
    )
    client._protocol = None

    Proxy_from_url_mock = mocker.patch('python_socks.async_.asyncio.Proxy.from_url', return_value=Mock(
        connect=AsyncMock(side_effect=python_socks.ProxyError('bad proxy connection')),
    ))
    mocker.patch.object(client, '_create_ssl_context')
    catch_connection_exceptions_mock = mocker.patch('aiobtclientrpc._utils.catch_connection_exceptions',
                                                    AsyncMock(return_value=('mock transport', 'mock protocol')))
    mocker.patch.object(client._loop, 'create_connection', Mock())
    mocker.patch.object(client, 'call', AsyncMock(side_effect=(
        'mock client_version',
        'mock login response',
    )))

    with pytest.raises(_errors.ConnectionError, match=r'^bad proxy connection$'):
        await client.login('Username', 'Password')

    assert Proxy_from_url_mock.call_args_list == [call(client._proxy_url.with_auth)]
    assert Proxy_from_url_mock.return_value.connect.call_args_list == [call(
        dest_host=client._host,
        dest_port=client._port,
        timeout=float('inf'),
    )]
    assert catch_connection_exceptions_mock.call_args_list == []
    assert client._loop.create_connection.call_args_list == []
    assert client.call.call_args_list == []

@pytest.mark.asyncio
async def test_DelugeRPCClient_login_without_proxy_url(mocker):
    client = _deluge._DelugeRPCClient(
        host='localhost',
        port=123,
    )
    client._protocol = None

    Proxy_from_url_mock = mocker.patch('python_socks.async_.asyncio.Proxy.from_url', return_value=Mock(
        connect=AsyncMock(),
    ))
    mocker.patch.object(client, '_create_ssl_context')
    catch_connection_exceptions_mock = mocker.patch('aiobtclientrpc._utils.catch_connection_exceptions',
                                                    AsyncMock(return_value=('mock transport', 'mock protocol')))
    mocker.patch.object(client._loop, 'create_connection', Mock())
    mocker.patch.object(client, 'call', AsyncMock(side_effect=(
        'mock client_version',
        'mock login response',
    )))

    await client.login('Username', 'Password')

    assert Proxy_from_url_mock.call_args_list == []
    assert Proxy_from_url_mock.return_value.connect.call_args_list == []
    assert catch_connection_exceptions_mock.call_args_list == [call(client._loop.create_connection.return_value)]
    assert client._loop.create_connection.call_args_list == [call(
        host=client._host,
        port=client._port,
        protocol_factory=client._protocol_factory,
        ssl=client._create_ssl_context.return_value,
        ssl_handshake_timeout=float('inf'),
    )]
    assert client.call.call_args_list == [
        call('daemon.info'),
        call('daemon.login', 'Username', 'Password', client_version='mock client_version'),
    ]


@pytest.mark.parametrize('protocol', (None, Mock()))
@pytest.mark.asyncio
async def test_DelugeRPCClient_logout(protocol, mocker):
    client = _deluge._DelugeRPCClient(host='localhost', port=123)
    mocker.patch.object(client, '_protocol', protocol)
    await client.logout()
    if protocol:
        assert client._protocol.close.call_args_list == [call()]


@pytest.mark.asyncio
async def test_DelugeRPCClient_call(mocker):
    client = _deluge._DelugeRPCClient(host='localhost', port=123)
    mocker.patch.object(client, '_protocol', Mock(send_request=AsyncMock()))
    return_value = await client.call('mock_method', 1, 2, 3, a='b')
    assert return_value is client._protocol.send_request.return_value
    exp_request = _deluge._DelugeRPCRequest(
        method='mock_method',
        args=(1, 2, 3),
        kwargs={'a': 'b'},
    )
    assert client._protocol.send_request.call_args_list == [call(exp_request)]


def test_DelugeRPCProtocol_initialization(mocker):
    on_connection_made = Mock()
    on_connection_lost = Mock()
    on_event_received = Mock()
    mocker.patch('aiobtclientrpc._deluge._DelugeRPCProtocol._reset_internal_state')
    protocol = _deluge._DelugeRPCProtocol(
        on_connection_made=on_connection_made,
        on_connection_lost=on_connection_lost,
        on_event_received=on_event_received,
    )
    assert protocol._on_connection_made is on_connection_made
    assert protocol._on_connection_lost is on_connection_lost
    assert protocol._on_event_received is on_event_received
    assert protocol._reset_internal_state.call_args_list == [call()]


def test_DelugeRPCProtocol_reset_internal_state(mocker):
    protocol = _deluge._DelugeRPCProtocol()
    protocol._transport = 'mock transport'
    protocol._buffer = b'partial messag'
    protocol._message_length = 123
    protocol._requests = {52: Mock(), 53: Mock()}
    _deluge._DelugeRPCRequest._next_id = 456

    protocol._reset_internal_state()

    assert protocol._transport is None
    assert protocol._buffer == b''
    assert protocol._message_length == 0
    assert protocol._requests == {}
    assert _deluge._DelugeRPCRequest._next_id == 0


@pytest.mark.parametrize('on_connection_made', (False, True))
def test_DelugeRPCProtocol_connection_made(on_connection_made, mocker):
    protocol = _deluge._DelugeRPCProtocol(
        on_connection_made=Mock() if on_connection_made else None,
    )
    transport = Mock()
    protocol.connection_made(transport)
    assert protocol._transport is transport
    if on_connection_made:
        assert protocol._on_connection_made.call_args_list == [call()]


@pytest.mark.parametrize('on_connection_lost', (False, True))
@pytest.mark.parametrize(
    argnames='exception, exp_exception',
    argvalues=(
        (None, _errors.ConnectionError('Connection lost')),
        (_errors.RPCError('error message'), _errors.RPCError('error message')),
    ),
    ids=lambda v: repr(v),
)
def test_DelugeRPCProtocol_connection_lost(exception, exp_exception, on_connection_lost, mocker):
    protocol = _deluge._DelugeRPCProtocol(
        on_connection_lost=Mock() if on_connection_lost else None,
    )
    mocker.patch.object(protocol, 'close')
    mocker.patch.object(protocol, '_reset_internal_state')
    _deluge._DelugeRPCRequest._next_id = 123
    finished_request = Mock(future=Mock(done=Mock(return_value=True)))
    unfinished_request = Mock(future=Mock(done=Mock(return_value=False)))
    protocol._requests = {
        23: finished_request,
        24: unfinished_request
    }

    protocol.connection_lost(exception)

    assert finished_request.future.set_exception.call_args_list == []
    assert finished_request.future.set_result.call_args_list == []
    assert unfinished_request.future.set_exception.call_args_list == [call(exp_exception)]
    assert unfinished_request.future.set_result.call_args_list == []
    assert protocol._requests == {}
    assert protocol.close.call_args_list == [call()]
    assert protocol._reset_internal_state.call_args_list == [call()]
    if on_connection_lost:
        assert protocol._on_connection_lost.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='transport, exp_calls',
    argvalues=(
        (None, []),
        (Mock(), [call()]),
    ),
)
def test_DelugeRPCProtocol_close(transport, exp_calls, mocker):
    protocol = _deluge._DelugeRPCProtocol()
    mocker.patch.object(protocol, '_transport', transport)
    protocol.close()
    if transport:
        assert protocol._transport.close.call_args_list == exp_calls
    else:
        assert protocol._transport is None


def test_DelugeRPCProtocol_data_received_ignores_short_buffer(mocker):
    protocol = _deluge._DelugeRPCProtocol()
    mocker.patch.object(protocol, '_handle_new_message')
    mocker.patch.object(protocol, '_handle_complete_message')

    protocol._buffer = b'x'
    protocol._message_length = 123
    protocol.data_received(b'')

    assert protocol._handle_new_message.call_args_list == []
    assert protocol._handle_complete_message.call_args_list == []
    assert protocol._message_length == 123
    assert protocol._buffer == b'x'

def test_DelugeRPCProtocol_data_received_reads_new_message(mocker):
    protocol = _deluge._DelugeRPCProtocol()
    mocker.patch.object(protocol, '_handle_new_message',
                        side_effect=lambda: setattr(protocol, '_message_length', 123))
    mocker.patch.object(protocol, '_handle_complete_message',
                        side_effect=lambda: setattr(protocol, '_buffer', b''))

    protocol._buffer = b'x' * protocol.MESSAGE_HEADER_SIZE
    protocol._message_length = 0
    protocol.data_received(b'more data')
    assert protocol._handle_new_message.call_args_list == [call()]
    assert protocol._handle_complete_message.call_args_list == []

def test_DelugeRPCProtocol_data_received_reads_complete_message(mocker):
    protocol = _deluge._DelugeRPCProtocol()
    mocker.patch.object(protocol, '_handle_new_message',
                        side_effect=lambda: setattr(protocol, '_message_length', 123))
    mocker.patch.object(protocol, '_handle_complete_message',
                        side_effect=lambda: setattr(protocol, '_buffer', b''))

    protocol._message_length = 33
    protocol._buffer = b'x' * 30
    protocol.data_received(b'more data')
    assert protocol._handle_new_message.call_args_list == []
    assert protocol._handle_complete_message.call_args_list == [call()]


def pack_message(body, protocol_version=_deluge._DelugeRPCProtocol.PROTOCOL_VERSION):
    return struct.pack(
        f'{_deluge._DelugeRPCProtocol.MESSAGE_HEADER_FORMAT}{len(body)}s',
        protocol_version,
        len(body),
        body,
    )


def test_DelugeRPCProtocol_handle_new_message_gets_unexpected_protocol_version(mocker):
    protocol = _deluge._DelugeRPCProtocol()
    body = b'foo'
    protocol._buffer = pack_message(body, protocol_version=29)
    with pytest.raises(RuntimeError, match='^Unsupported protocol version: 29$'):
        protocol._handle_new_message()

def test_DelugeRPCProtocol_handle_new_message_consumes_header(mocker):
    protocol = _deluge._DelugeRPCProtocol()
    body = b'foo'
    protocol._buffer = pack_message(body, protocol_version=protocol.PROTOCOL_VERSION)
    protocol._handle_new_message()
    assert protocol._message_length == len(body)
    assert protocol._buffer == body


@pytest.mark.parametrize('on_event_received', (None, Mock()))
def test_DelugeRPCProtocol_handle_complete_message_gets_RPC_EVENT(on_event_received, mocker):
    protocol = _deluge._DelugeRPCProtocol(
        on_event_received=on_event_received,
    )
    mocker.patch.object(protocol, '_create_rpc_error')
    mocker.patch.object(protocol, '_set_response')

    args = ('arg1', 'órg2', 'ärg3')
    msg = (protocol.RPC_EVENT, 'MockEventName', args)
    data = zlib.compress(rencode.dumps(msg))
    protocol._buffer = data + b'remaining data'
    protocol._message_length = len(data)

    protocol._handle_complete_message()

    if on_event_received:
        assert protocol._on_event_received.call_args_list == [call('MockEventName', args)]
    else:
        assert protocol._on_event_received is None
    assert protocol._create_rpc_error.call_args_list == []
    assert protocol._set_response.call_args_list == []
    assert protocol._buffer == b'remaining data'
    assert protocol._message_length == 0

def test_DelugeRPCProtocol_handle_complete_message_gets_RPC_ERROR(mocker):
    protocol = _deluge._DelugeRPCProtocol(
        on_event_received=Mock(),
    )
    mocker.patch.object(protocol, '_create_rpc_error')
    mocker.patch.object(protocol, '_set_response')

    request_id = 123
    exc_clsname = 'MockError'
    exc_posargs = ('arg1', 'órg2', 'ärg3')
    exc_kwargs = {'foo': 'bar'}
    traceback = 'mock traceback'
    msg = (protocol.RPC_ERROR, request_id, exc_clsname, exc_posargs, exc_kwargs, traceback)
    data = zlib.compress(rencode.dumps(msg))
    protocol._buffer = data + b'remaining data'
    protocol._message_length = len(data)

    protocol._handle_complete_message()

    assert protocol._create_rpc_error.call_args_list == [call(exc_clsname, exc_posargs, exc_kwargs, traceback)]
    assert protocol._set_response.call_args_list == [call(request_id, protocol._create_rpc_error.return_value)]
    assert protocol._on_event_received.call_args_list == []
    assert protocol._buffer == b'remaining data'
    assert protocol._message_length == 0

def test_DelugeRPCProtocol_handle_complete_message_gets_RPC_RESPONSE(mocker):
    protocol = _deluge._DelugeRPCProtocol(
        on_event_received=Mock(),
    )
    mocker.patch.object(protocol, '_create_rpc_error')
    mocker.patch.object(protocol, '_set_response')

    request_id = 123
    response = ('arg1', 'órg2', 'ärg3')
    msg = (protocol.RPC_RESPONSE, request_id, response)
    data = zlib.compress(rencode.dumps(msg))
    protocol._buffer = data + b'remaining data'
    protocol._message_length = len(data)

    protocol._handle_complete_message()

    assert protocol._create_rpc_error.call_args_list == []
    assert protocol._set_response.call_args_list == [call(request_id, response)]
    assert protocol._on_event_received.call_args_list == []
    assert protocol._buffer == b'remaining data'
    assert protocol._message_length == 0

def test_DelugeRPCProtocol_handle_complete_message_gets_unknown_msg_type(mocker):
    protocol = _deluge._DelugeRPCProtocol(
        on_event_received=Mock(),
    )
    mocker.patch.object(protocol, '_create_rpc_error')
    mocker.patch.object(protocol, '_set_response')

    request_id = 123
    response = ('arg1', 'órg2', 'ärg3')
    msg = ('gnarf', request_id, response)
    data = zlib.compress(rencode.dumps(msg))
    protocol._buffer = data + b'remaining data'
    protocol._message_length = len(data)

    with pytest.raises(RuntimeError, match="^Unknown RPC message type: 'gnarf'$"):
        protocol._handle_complete_message()

    assert protocol._create_rpc_error.call_args_list == []
    assert protocol._set_response.call_args_list == []
    assert protocol._on_event_received.call_args_list == []
    assert protocol._buffer == b'remaining data'
    assert protocol._message_length == 0


@pytest.mark.parametrize(
    argnames='clsname, posargs, kwargs, traceback, exp_exception',
    argvalues=(
        ('BadLoginError', (), {}, 'mock traceback',
         _errors.AuthenticationError('Authentication failed')),
        ('NotAuthorizedError', (), {}, 'mock traceback',
         _errors.AuthenticationError('Not authorized')),
        ('NotAuthorizedError', (5, 7), {}, 'mock traceback',
         _errors.AuthenticationError('Not authorized: Your authorization level is 5, but you need 7')),
        ('AnyError', ('An error message',), {}, 'mock traceback',
         _errors.RPCError('An error message')),
        ('AnyError', (), {}, 'mock traceback',
         _errors.RPCError('AnyError')),
        ('', (), {}, 'mock traceback',
         _errors.RPCError('Unknown error')),
    ),
)
def test_DelugeRPCProtocol_create_rpc_error(clsname, posargs, kwargs, traceback, exp_exception):
    protocol = _deluge._DelugeRPCProtocol()
    exc = protocol._create_rpc_error(clsname, posargs, kwargs, traceback)
    assert type(exc) is type(exp_exception)
    assert str(exc) == str(exp_exception)


def test_DelugeRPCProtocol_set_response_fails_to_find_request_id():
    protocol = _deluge._DelugeRPCProtocol()
    response = 'mock response'
    assert '123' not in protocol._requests
    with pytest.raises(RuntimeError, match=rf"^Got response to unknown request #123: {response!r}$"):
        protocol._set_response('123', response)
    assert '123' not in protocol._requests

def test_DelugeRPCProtocol_set_response_gets_request_with_finished_future():
    protocol = _deluge._DelugeRPCProtocol()
    request = Mock(
        id=123,
        future=Mock(done=Mock(return_value=True)),
    )
    response = 'mock response'
    protocol._requests[request.id] = request
    with pytest.raises(RuntimeError, match=r'^Request #123 already has a response$'):
        protocol._set_response(request.id, response)
    assert request.future.set_exception.call_args_list == []
    assert request.future.set_result.call_args_list == []
    assert request.id not in protocol._requests

def test_DelugeRPCProtocol_set_response_sets_exception():
    protocol = _deluge._DelugeRPCProtocol()
    request = Mock(
        id=123,
        future=Mock(done=Mock(return_value=False)),
    )
    response = TypeError('This is the error message')
    protocol._requests[request.id] = request
    protocol._set_response(request.id, response)
    assert request.future.set_exception.call_args_list == [call(response)]
    assert request.future.set_result.call_args_list == []
    assert request.id not in protocol._requests

def test_DelugeRPCProtocol_set_response_sets_result():
    protocol = _deluge._DelugeRPCProtocol()
    request = Mock(
        id=123,
        future=Mock(done=Mock(return_value=False)),
    )
    response = 'This is the response'
    protocol._requests[request.id] = request
    protocol._set_response(request.id, response)
    assert request.future.set_exception.call_args_list == []
    assert request.future.set_result.call_args_list == [call(response)]
    assert request.id not in protocol._requests

def test_DelugeRPCProtocol_set_response_removes_request_on_unexpected_exception():
    protocol = _deluge._DelugeRPCProtocol()
    request = Mock(
        id=123,
        future=Mock(done=Mock(side_effect=ValueError('Oops'))),
    )
    response = 'This is the response'
    protocol._requests[request.id] = request
    with pytest.raises(ValueError, match=r'^Oops$'):
        protocol._set_response(request.id, response)
    assert request.future.set_exception.call_args_list == []
    assert request.future.set_result.call_args_list == []
    assert request.id not in protocol._requests


def test_DelugeRPCProtocol_transfer_messages(mocker):
    compress_mock = mocker.patch('zlib.compress', return_value=b'compressed data')
    dumps_mock = mocker.patch('rencode.dumps', return_value=b'dumped data')
    convert_to_basic_type_mock = mocker.patch(
        'aiobtclientrpc._utils.convert_to_basic_type',
        return_value=b'converted data',
    )
    data = 'mock data'
    protocol = _deluge._DelugeRPCProtocol()
    mocker.patch.object(protocol, '_transport')
    protocol._transfer_messages(data)
    assert protocol._transport.write.call_args_list == [call(pack_message(b'compressed data'))]
    assert compress_mock.call_args_list == [call(b'dumped data')]
    assert dumps_mock.call_args_list == [call(b'converted data')]
    assert convert_to_basic_type_mock.call_args_list == [call(('mock data',))]


def test_DelugeRPCProtocol_send_request(mocker):
    protocol = _deluge._DelugeRPCProtocol()
    mocker.patch.object(protocol, '_transfer_messages')
    request = Mock(id=123)
    return_value = protocol.send_request(request)
    assert protocol._requests[request.id] is request
    assert protocol._transfer_messages.call_args_list == [call(request.format_message())]
    assert return_value is request.future


@pytest.mark.asyncio
async def test_DelugeRPCRequest():
    _deluge._DelugeRPCRequest.reset_id()
    for _ in range(3):
        for i in range(10):
            request = _deluge._DelugeRPCRequest(
                method='method.name',
                args=['arg1', 'arg2'],
                kwargs=(('foo', 'bar'), ('baz', 'arf')),
            )
            assert request.id == i
            assert request.args == ('arg1', 'arg2')
            assert request.kwargs == {'foo': 'bar', 'baz': 'arf'}
            assert asyncio.isfuture(request.future)
            assert request.format_message() == (
                request.id,
                request.method,
                request.args,
                request.kwargs,
            )
        _deluge._DelugeRPCRequest.reset_id()

@pytest.mark.parametrize('kwargs', ({}, {'foo': 'bar'}, {'foo': 'bar', 'baz': 'arf'}))
@pytest.mark.parametrize('args', ((), ('arg1',), ('arg1', 'arg2')))
@pytest.mark.asyncio
async def test_DelugeRPCRequest_repr(kwargs, args):
    request = _deluge._DelugeRPCRequest(
        method='method.name',
        args=args,
        kwargs=kwargs,
    )
    exp_repr = f'{request.method}('
    if args:
        exp_repr += ', '.join((f'{a!r}' for a in request.args))
    if kwargs:
        if args:
            exp_repr += ', '
        exp_repr += ', '.join((f'{k}={v!r}' for k, v in request.kwargs.items()))
    exp_repr += ')'
    assert repr(request) == exp_repr

@pytest.mark.parametrize(
    argnames='a, b, exp_equal',
    argvalues=(
        (
            {'method': 'method1', 'args': (1, 2, 3), 'kwargs': {'foo': 'bar'}},
            {'method': 'method1', 'args': (1, 2, 3), 'kwargs': {'foo': 'bar'}},
            True,
        ),
        (
            {'method': 'method1', 'args': (1, 2, 3), 'kwargs': {'foo': 'bar'}},
            {'method': 'method2', 'args': (1, 2, 3), 'kwargs': {'foo': 'bar'}},
            False,
        ),
        (
            {'method': 'method1', 'args': (1, 2, 3), 'kwargs': {'foo': 'bar'}},
            {'method': 'method1', 'args': (1, 2), 'kwargs': {'foo': 'bar'}},
            False,
        ),
        (
            {'method': 'method1', 'args': (1, 2, 3), 'kwargs': {'foo': 'bar'}},
            {'method': 'method1', 'args': (1, 2, 3), 'kwargs': {'foo': 'Bar'}},
            False,
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_DelugeRPCRequest_equality(a, b, exp_equal, event_loop):
    if exp_equal:
        assert _deluge._DelugeRPCRequest(**a) == _deluge._DelugeRPCRequest(**b)
    else:
        assert _deluge._DelugeRPCRequest(**a) != _deluge._DelugeRPCRequest(**b)



import asyncio
import re
from unittest.mock import Mock, PropertyMock, call

import httpx
import httpx_socks
import pytest

from aiobtclientrpc import __project_name__, __version__, _errors, _utils

from .common import AsyncMock, make_url_parts


@pytest.mark.asyncio
async def test_get_aioloop_with_running_loop(mocker):
    loop = _utils.get_aioloop()
    assert loop.is_running()
    assert isinstance(loop, asyncio.AbstractEventLoop)

def test_get_aioloop_without_running_loop(mocker):
    loop = _utils.get_aioloop()
    assert isinstance(loop, asyncio.AbstractEventLoop)


def test_clients(mocker):
    import aiobtclientrpc  # isort:skip
    assert _utils.clients() == [
        aiobtclientrpc.DelugeRPC,
        aiobtclientrpc.QbittorrentRPC,
        aiobtclientrpc.RtorrentRPC,
        aiobtclientrpc.TransmissionRPC,
    ]


@pytest.mark.parametrize(
    argnames='names, name, args, kwargs, exp_exception',
    argvalues=(
        (('foo', 'bar', 'baz'), 'foo', (1, 2, 3), {'hey': 'ho'}, None),
        (('foo', 'bar', 'baz'), 'bar', (1, 2, 3), {}, None),
        (('foo', 'bar', 'baz'), 'baz', (), {'hey': 'ho'}, None),
        (('foo', 'bar', 'baz'), 'asdf', (), {}, _errors.ValueError('No such client: asdf')),
    ),
)
def test_client(names, name, args, kwargs, exp_exception, mocker):
    def MockRPC(name):
        cls_mock = Mock()
        cls_mock.configure_mock(name=name)
        return cls_mock

    client_clses = [MockRPC(name) for name in names]
    mocker.patch('aiobtclientrpc._utils.clients', return_value=client_clses)

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            _utils.client(name, *args, **kwargs)
        for cls in client_clses:
            assert cls.call_args_list == []

    else:
        return_value = _utils.client(name, *args, **kwargs)
        for cls in client_clses:
            if cls.name == name:
                assert return_value is cls.return_value
                assert cls.call_args_list == [call(*args, **kwargs)]
            else:
                assert return_value is not cls.return_value
                assert cls.call_args_list == []


def test_ConnectionStatus():
    assert _utils.ConnectionStatus('connecting') == _utils.ConnectionStatus.connecting
    assert _utils.ConnectionStatus('connected') == _utils.ConnectionStatus.connected
    assert _utils.ConnectionStatus('disconnected') == _utils.ConnectionStatus.disconnected


def test_cached_property():
    expensive_calculation = Mock(return_value='expensive value')

    class Foo:
        @_utils.cached_property
        def bar(self):
            return expensive_calculation('a', 'b', c='see')

    foo = Foo()
    for _ in range(3):
        assert foo.bar == 'expensive value'
        assert expensive_calculation.call_args_list == [call('a', 'b', c='see')]


@pytest.mark.parametrize(
    argnames='url, exp_parts_or_exception',
    argvalues=(
        (None, {'scheme': None, 'host': None, 'port': None, 'path': None, 'username': None, 'password': None}),
        ('', {'scheme': None, 'host': None, 'port': None, 'path': None, 'username': None, 'password': None}),

        # No scheme without path
        ('localhost',
         {'scheme': None, 'host': 'localhost', 'port': None, 'path': None, 'username': None, 'password': None}),
        ('localhost:123',
         {'scheme': None, 'host': 'localhost', 'port': '123', 'path': None, 'username': None, 'password': None}),
        ('foo:bar@localhost:123',
         {'scheme': None, 'host': 'localhost', 'port': '123', 'path': None, 'username': 'foo', 'password': 'bar'}),
        ('foo:@localhost:123',
         {'scheme': None, 'host': 'localhost', 'port': '123', 'path': None, 'username': 'foo', 'password': None}),
        (':bar@localhost:123',
         {'scheme': None, 'host': 'localhost', 'port': '123', 'path': None, 'username': None, 'password': 'bar'}),
        ('localhost:arf',
         _errors.ValueError('Invalid port')),

        # No scheme with path
        ('localhost/some/path',
         {'scheme': None, 'host': 'localhost', 'port': None, 'path': '/some/path', 'username': None, 'password': None}),
        ('localhost:123/some/path',
         {'scheme': None, 'host': 'localhost', 'port': '123', 'path': '/some/path', 'username': None, 'password': None}),
        ('foo:bar@localhost:123/some/path',
         {'scheme': None, 'host': 'localhost', 'port': '123', 'path': '/some/path', 'username': 'foo', 'password': 'bar'}),
        ('foo:@localhost:123/some/path',
         {'scheme': None, 'host': 'localhost', 'port': '123', 'path': '/some/path', 'username': 'foo', 'password': None}),
        (':bar@localhost:123/some/path',
         {'scheme': None, 'host': 'localhost', 'port': '123', 'path': '/some/path', 'username': None, 'password': 'bar'}),
        ('localhost:arf/some/path',
         _errors.ValueError('Invalid port')),

        # Scheme without path
        ('ftp://localhost',
         {'scheme': 'ftp', 'host': 'localhost', 'port': None, 'path': None, 'username': None, 'password': None}),
        ('ftp://localhost:123',
         {'scheme': 'ftp', 'host': 'localhost', 'port': '123', 'path': None, 'username': None, 'password': None}),
        ('ftp://foo:bar@localhost:123',
         {'scheme': 'ftp', 'host': 'localhost', 'port': '123', 'path': None, 'username': 'foo', 'password': 'bar'}),
        ('ftp://foo:@localhost:123',
         {'scheme': 'ftp', 'host': 'localhost', 'port': '123', 'path': None, 'username': 'foo', 'password': None}),
        ('ftp://:bar@localhost:123',
         {'scheme': 'ftp', 'host': 'localhost', 'port': '123', 'path': None, 'username': None, 'password': 'bar'}),
        ('ftp://localhost:arf',
         _errors.ValueError('Invalid port')),

        # Scheme with path
        ('http://localhost/some/path',
         {'scheme': 'http', 'host': 'localhost', 'port': None, 'path': '/some/path', 'username': None, 'password': None}),
        ('http://localhost:123/some/path',
         {'scheme': 'http', 'host': 'localhost', 'port': '123', 'path': '/some/path', 'username': None, 'password': None}),
        ('http://foo:bar@localhost:123/some/path',
         {'scheme': 'http', 'host': 'localhost', 'port': '123', 'path': '/some/path', 'username': 'foo', 'password': 'bar'}),
        ('http://foo:@localhost:123/some/path',
         {'scheme': 'http', 'host': 'localhost', 'port': '123', 'path': '/some/path', 'username': 'foo', 'password': None}),
        ('http://:bar@localhost:123/some/path',
         {'scheme': 'http', 'host': 'localhost', 'port': '123', 'path': '/some/path', 'username': None, 'password': 'bar'}),
        ('http://localhost:arf/some/path',
         _errors.ValueError('Invalid port')),

        # File system path
        ('file://relative/path',
         {'scheme': 'file', 'host': None, 'port': None, 'path': 'relative/path', 'username': None, 'password': None}),
        ('file:///absolute/path',
         {'scheme': 'file', 'host': None, 'port': None, 'path': '/absolute/path', 'username': None, 'password': None}),
        ('file://foo:bar@localhost',
         {'scheme': 'file', 'host': None, 'port': None, 'path': 'foo:bar@localhost', 'username': None, 'password': None}),
        ('file://localhost:123',
         {'scheme': 'file', 'host': None, 'port': None, 'path': 'localhost:123', 'username': None, 'password': None}),
        ('file://localhost:arf',
         {'scheme': 'file', 'host': None, 'port': None, 'path': 'localhost:arf', 'username': None, 'password': None}),
        ('/absolute/path',
         {'scheme': 'file', 'host': None, 'port': None, 'path': '/absolute/path', 'username': None, 'password': None}),
    ),
    ids=lambda v: str(v),
)
def test_URL_parsing(url, exp_parts_or_exception):
    def make_url(url):
        if url is not None:
            return _utils.URL(url)
        else:
            return _utils.URL()

    if isinstance(exp_parts_or_exception, Exception):
        exception = exp_parts_or_exception
        with pytest.raises(type(exception), match=rf'^{re.escape(str(exception))}$'):
            make_url(url)
    else:
        url = make_url(url)
        exp_parts = exp_parts_or_exception
        parts = make_url_parts(url)
        assert parts == exp_parts

default_url_testcases = (
    (
        'foo:bar@localhost:123/some/path',
        'socks://x:y@defaulthost:999/default/path',
        {'scheme': 'socks', 'host': 'localhost', 'port': '123', 'path': '/some/path', 'username': 'foo', 'password': 'bar'},
    ),
    (
        'http://:bar@localhost:123/some/path',
        'socks://x:y@defaulthost:999/default/path',
        {'scheme': 'http', 'host': 'localhost', 'port': '123', 'path': '/some/path', 'username': 'x', 'password': 'bar'},
    ),
    (
        'http://foo:@localhost:123/some/path',
        'socks://x:y@defaulthost:999/default/path',
        {'scheme': 'http', 'host': 'localhost', 'port': '123', 'path': '/some/path', 'username': 'foo', 'password': 'y'},
    ),
    (
        'http://foo:bar@localhost/some/path',
        'socks://x:y@defaulthost:999/default/path',
        {'scheme': 'http', 'host': 'localhost', 'port': '999', 'path': '/some/path', 'username': 'foo', 'password': 'bar'},
    ),
    (
        'http://foo:bar@localhost:123/',
        'socks://x:y@defaulthost:999/default/path',
        {'scheme': 'http', 'host': 'localhost', 'port': '123', 'path': '/', 'username': 'foo', 'password': 'bar'},
    ),
    (
        'http://foo:bar@localhost:123',
        'socks://x:y@defaulthost:999/default/path',
        {'scheme': 'http', 'host': 'localhost', 'port': '123', 'path': '/default/path', 'username': 'foo', 'password': 'bar'},
    ),
    (
        'http://foo:bar@localhost:asdf',
        None,
        _errors.ValueError('Invalid port'),
    ),
)

@pytest.mark.parametrize('url, default_url, exp_parts_or_exception', default_url_testcases)
def test_URL_default_url_attribute(url, default_url, exp_parts_or_exception):
    class MyURL(_utils.URL):
        default = default_url

    assert isinstance(MyURL.default, MyURL)

    if isinstance(exp_parts_or_exception, Exception):
        exception = exp_parts_or_exception
        with pytest.raises(type(exception), match=rf'^{re.escape(str(exception))}$'):
            MyURL(url)
    else:
        url = MyURL(url)
        exp_parts = exp_parts_or_exception
        parts = make_url_parts(url)
        assert parts == exp_parts

@pytest.mark.parametrize('url, default_url, exp_parts_or_exception', default_url_testcases)
def test_URL_default_url_argument(url, default_url, exp_parts_or_exception):
    class MyURL(_utils.URL):
        default = 'not://the:correct@url:111/to/use'

    if isinstance(exp_parts_or_exception, Exception):
        exception = exp_parts_or_exception
        with pytest.raises(type(exception), match=rf'^{re.escape(str(exception))}$'):
            MyURL(url, default=default_url)
    else:
        url = MyURL(url, default=default_url)
        exp_parts = exp_parts_or_exception
        parts = make_url_parts(url)
        assert parts == exp_parts


@pytest.mark.parametrize(
    argnames='url, new_scheme, exp_before, exp_after',
    argvalues=(
        ('this://a:b@localhost:555/some/path', 'that',
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'that', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', 'FILE',
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'file', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path','Socks5',
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'socks5', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', 123,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': '123', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', False,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': None, 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
    ),
)
def test_URL_scheme(url, new_scheme, exp_before, exp_after):
    cb = Mock()
    url = _utils.URL(url, on_change=cb)
    assert cb.call_args_list == []
    assert make_url_parts(url) == exp_before
    url.scheme = new_scheme
    assert make_url_parts(url) == exp_after
    assert cb.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='url, new_host, exp_before, exp_after',
    argvalues=(
        ('this://a:b@localhost:555/some/path', '127.0.0.1',
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': '127.0.0.1', 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', 123,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': '123', 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', None,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': None, 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', False,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': None, 'port': '555', 'path': '/some/path'}),
    ),
)
def test_URL_host(url, new_host, exp_before, exp_after):
    cb = Mock()
    url = _utils.URL(url, on_change=cb)
    assert cb.call_args_list == []
    assert make_url_parts(url) == exp_before
    url.host = new_host
    assert make_url_parts(url) == exp_after
    assert cb.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='url, new_port, exp_before, exp_after',
    argvalues=(
        ('this://a:b@localhost:555/some/path', 0,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': None, 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', 1,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '1', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', 65535,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '65535', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', -1,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         _errors.ValueError('Invalid port')),
        ('this://a:b@localhost:555/some/path', '-1',
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         _errors.ValueError('Invalid port')),
        ('this://a:b@localhost:555/some/path', 65536,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         _errors.ValueError('Invalid port')),
        ('this://a:b@localhost:555/some/path', 'foo',
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         _errors.ValueError('Invalid port')),
        ('this://a:b@localhost:555/some/path', None,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': None, 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', False,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': None, 'path': '/some/path'}),
    ),
)
def test_URL_port(url, new_port, exp_before, exp_after):
    cb = Mock()
    url = _utils.URL(url, on_change=cb)
    original_port = url.port
    assert cb.call_args_list == []
    assert make_url_parts(url) == exp_before
    if isinstance(exp_after, Exception):
        with pytest.raises(type(exp_after), match=rf'^{re.escape(str(exp_after))}$'):
            url.port = new_port
        assert url.port == original_port
        assert cb.call_args_list == []
    else:
        url.port = new_port
        assert make_url_parts(url) == exp_after
        assert cb.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='url, new_path, exp_before, exp_after',
    argvalues=(
        ('this://a:b@localhost:555/some/path', 'other/path',
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': 'other/path'}),
        ('this://a:b@localhost:555/some/path', '/other/path',
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/other/path'}),
        ('this://a:b@localhost:555/some/path', 123,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '123'}),
        ('this://a:b@localhost:555/some/path', None,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': None}),
        ('this://a:b@localhost:555/some/path', False,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': None}),
    ),
)
def test_URL_path(url, new_path, exp_before, exp_after):
    cb = Mock()
    url = _utils.URL(url, on_change=cb)
    assert cb.call_args_list == []
    assert make_url_parts(url) == exp_before
    url.path = new_path
    assert make_url_parts(url) == exp_after
    assert cb.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='url, new_username, exp_before, exp_after',
    argvalues=(
        ('this://a:b@localhost:555/some/path', 'x',
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'x', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', 123,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': '123', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', None,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': None, 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', False,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': None, 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
    ),
)
def test_URL_username(url, new_username, exp_before, exp_after):
    cb = Mock()
    url = _utils.URL(url, on_change=cb)
    assert cb.call_args_list == []
    assert make_url_parts(url) == exp_before
    url.username = new_username
    assert make_url_parts(url) == exp_after
    assert cb.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='url, new_password, exp_before, exp_after',
    argvalues=(
        ('this://a:b@localhost:555/some/path', 'x',
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': 'x', 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', 123,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': '123', 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', None,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': None, 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
        ('this://a:b@localhost:555/some/path', False,
         {'scheme': 'this', 'username': 'a', 'password': 'b', 'host': 'localhost', 'port': '555', 'path': '/some/path'},
         {'scheme': 'this', 'username': 'a', 'password': None, 'host': 'localhost', 'port': '555', 'path': '/some/path'}),
    ),
)
def test_URL_password(url, new_password, exp_before, exp_after):
    cb = Mock()
    url = _utils.URL(url, on_change=cb)
    assert cb.call_args_list == []
    assert make_url_parts(url) == exp_before
    url.password = new_password
    assert make_url_parts(url) == exp_after
    assert cb.call_args_list == [call()]


@pytest.mark.parametrize('path, exp_path', (('', ''), ('/some/path', '/some/path')))
@pytest.mark.parametrize('port, exp_port', (('', ''), (':123', ':123')))
@pytest.mark.parametrize(
    argnames='url, exp_url',
    argvalues=(
        ('http://a:b@localhost{port}{path}', 'http://localhost{exp_port}{exp_path}'),
        ('http://:b@localhost{port}{path}', 'http://localhost{exp_port}{exp_path}'),
        ('http://a:@localhost{port}{path}', 'http://localhost{exp_port}{exp_path}'),
        ('http://:@localhost{port}{path}', 'http://localhost{exp_port}{exp_path}'),
        ('http://localhost{port}{path}', 'http://localhost{exp_port}{exp_path}'),

        # For file:// URLs, "username:password@" has no special meaning
        ('file://a:b@localhost{port}{path}', 'file://a:b@localhost{exp_port}{exp_path}'),
        ('file://:b@localhost{port}{path}', 'file://:b@localhost{exp_port}{exp_path}'),
        ('file://a:@localhost{port}{path}', 'file://a:@localhost{exp_port}{exp_path}'),
        ('file://:@localhost{port}{path}', 'file://:@localhost{exp_port}{exp_path}'),
        ('file://localhost{port}{path}', 'file://localhost{exp_port}{exp_path}'),
    ),
)
def test_URL_without_auth(url, exp_url, port, exp_port, path, exp_path):
    url = _utils.URL(url.format(port=port, path=path))
    url_without_auth = url.without_auth
    exp_url_without_auth = exp_url.format(exp_port=exp_port, exp_path=exp_path)
    assert url_without_auth == exp_url_without_auth


@pytest.mark.parametrize('path, exp_path', (('', ''), ('/some/path', '/some/path')))
@pytest.mark.parametrize('port, exp_port', (('', ''), (':123', ':123')))
@pytest.mark.parametrize(
    argnames='url, exp_url',
    argvalues=(
        ('http://a:b@localhost{port}{path}', 'http://a:b@localhost{exp_port}{exp_path}'),
        ('http://:b@localhost{port}{path}', 'http://:b@localhost{exp_port}{exp_path}'),
        ('http://a:@localhost{port}{path}', 'http://a:@localhost{exp_port}{exp_path}'),
        ('http://:@localhost{port}{path}', 'http://localhost{exp_port}{exp_path}'),
        ('http://localhost{port}{path}', 'http://localhost{exp_port}{exp_path}'),

        # For file:// URLs, "username:password@" has no special meaning
        ('file://a:b@localhost{port}{path}', 'file://a:b@localhost{exp_port}{exp_path}'),
        ('file://:b@localhost{port}{path}', 'file://:b@localhost{exp_port}{exp_path}'),
        ('file://a:@localhost{port}{path}', 'file://a:@localhost{exp_port}{exp_path}'),
        ('file://:@localhost{port}{path}', 'file://:@localhost{exp_port}{exp_path}'),
        ('file://localhost{port}{path}', 'file://localhost{exp_port}{exp_path}'),
    ),
)
def test_URL_with_auth(url, exp_url, port, exp_port, path, exp_path):
    url = _utils.URL(url.format(port=port, path=path))
    url_with_auth = url.with_auth
    exp_url_with_auth = exp_url.format(exp_port=exp_port, exp_path=exp_path)
    assert url_with_auth == exp_url_with_auth


@pytest.mark.parametrize(
    argnames='url1, url2, exp_equal',
    argvalues=(
        ('http://a:b@localhost:1234/some/path', 'http://a:b@localhost:1234/some/path', True),
        ('http://a:b@localhost:1234/some/path', 'http://a:b@localhost:1234/other/path', False),
        ('http://a:b@localhost:1234/some/path', 'http://a:b@localhost:1235/some/path', False),
        ('http://a:b@localhost:1234/some/path', 'http://a:b@localhoft:1234/some/path', False),
        ('http://a:b@localhost:1234/some/path', 'http://a:c@localhost:1234/some/path', False),
        ('http://a:b@localhost:1234/some/path', 'http://c:b@localhost:1234/some/path', False),
        ('http://a:b@localhost:1234/some/path', 'ftp://a:b@localhost:1234/some/path', False),
        ('http://a:b@localhost:1234/some/path', Mock(), NotImplemented),
    ),
)
def test_URL_equality(url1, url2, exp_equal, mocker):
    if exp_equal is True:
        assert _utils.URL(url1) == _utils.URL(url2)
        assert _utils.URL(url1) != url2
        assert url1 != _utils.URL(url2)
    elif exp_equal is False:
        assert _utils.URL(url1) != _utils.URL(url2)
        assert _utils.URL(url1) != url2
        assert url1 != _utils.URL(url2)
    else:
        assert _utils.URL(url1).__eq__(url2) is NotImplemented
        assert _utils.URL(url1).__ne__(url2) is NotImplemented


def test_URL_str(mocker):
    url = _utils.URL('this://localhost')
    mocker.patch.object(type(url), 'without_auth', PropertyMock(return_value='mocked URL'))
    assert str(url) == 'mocked URL'

@pytest.mark.parametrize('on_change', (None, Mock()))
def test_URL_repr_without_on_change_callback(on_change, mocker):
    url = _utils.URL('this://localhost', on_change=on_change)
    mocker.patch.object(type(url), 'without_auth', PropertyMock(return_value='mocked URL'))
    if on_change:
        assert repr(url) == f"URL('mocked URL', on_change={on_change!r})"
    else:
        assert repr(url) == "URL('mocked URL')"


@pytest.mark.parametrize('proxy_url', (None, 'socks5://foo.bar'))
@pytest.mark.parametrize(
    argnames='username, password',
    argvalues=(
        (None, None),
        ('', ''),
        ('foo', None),
        ('', 'bar'),
        ('foo', 'bar'),
    ),
)
def test_create_http_client(username, password, proxy_url, mocker):
    AsyncClient_mock = mocker.patch('httpx.AsyncClient')
    BasicAuth_mock = mocker.patch('httpx.BasicAuth')
    AsyncProxyTransport_mock = mocker.patch('httpx_socks.AsyncProxyTransport')

    client = _utils.create_http_client(auth=(username, password), proxy_url=proxy_url)
    assert client is AsyncClient_mock.return_value

    exp_AsyncClient_kwargs = {
        'timeout': float('inf'),
        'headers': {'User-Agent': f'{__project_name__} {__version__}'},
    }
    if username and password:
        assert BasicAuth_mock.call_args_list == [call(username, password)]
        exp_AsyncClient_kwargs['auth'] = BasicAuth_mock.return_value

    if proxy_url:
        assert AsyncProxyTransport_mock.from_url.call_args_list == [call(proxy_url)]
        exp_AsyncClient_kwargs['transport'] = AsyncProxyTransport_mock.from_url.return_value

    assert AsyncClient_mock.call_args_list == [call(**exp_AsyncClient_kwargs)]


@pytest.mark.parametrize(
    argnames='raised_exception, exp_exception',
    argvalues=(
        (httpx.HTTPError('Fail'), _errors.ConnectionError('Fail')),
        (httpx.HTTPError('[Errno 123] Ugly fail'), _errors.ConnectionError('Ugly fail')),
        (httpx.HTTPError('[Errno -123] Ugly negative fail'), _errors.ConnectionError('Ugly negative fail')),
        (httpx_socks.ProxyError('Fail'), _errors.ConnectionError('Fail')),
        (httpx_socks.ProxyError('[Errno 456] Ugly fail'), _errors.ConnectionError('Ugly fail')),
        (httpx_socks.ProxyError('[Errno -456] Ugly negative fail'), _errors.ConnectionError('Ugly negative fail')),
        (ConnectionAbortedError(), _errors.ConnectionError('Connection aborted')),
        (ConnectionRefusedError(), _errors.ConnectionError('Connection refused')),
        (ConnectionResetError(), _errors.ConnectionError('Connection reset')),
        (OSError(123, 'Fail'), _errors.ConnectionError('Fail')),
        (OSError('Fail'), _errors.ConnectionError('Fail')),
        (OSError(), _errors.ConnectionError('Unknown error')),
        (OSError("[Errno 123] Error message ('127.0.0.1', 345, 0, 0)"),
         _errors.ConnectionError('Error message')),
        (OSError("Multiple exceptions: [Errno 123] Error message ('::1', 456, 0, 0), "
                 "[Errno 123] Error message ('127.0.0.1', 456)"),
         _errors.ConnectionError('Error message')),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_catch_connection_exceptions(raised_exception, exp_exception, mocker):
    coro_function = AsyncMock(side_effect=raised_exception)
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await _utils.catch_connection_exceptions(coro_function())
    else:
        return_value = await _utils.catch_connection_exceptions(coro_function())
        assert return_value is coro_function.return_value


# We don't test `bool` here because it can't be subclassed.
MyBytearray = type('MyBytearray', (bytearray,), {})
MyBytes = type('MyBytes', (bytes,), {})
MyDict = type('MyDict', (dict,), {})
MyFloat = type('MyFloat', (float,), {})
MyInt = type('MyInt', (int,), {})
MyList = type('MyList', (list,), {})
MyStr = type('MyStr', (str,), {})
MyTuple = type('MyTuple', (tuple,), {})

@pytest.mark.parametrize(
    argnames='data, exp_result',
    argvalues=(
        (
            MyList([
                MyStr('one'),
                MyTuple([
                    'a',
                    MyDict({MyFloat(99.9): MyList([1, 2, 3])}),
                    MyInt(123),
                ]),
                MyBytes(b'asdf'),
            ]),
            [
                'one',
                (
                    'a',
                    {99.9: [1, 2, 3]},
                    123,
                ),
                b'asdf',
            ],
        ),
        (
            MyTuple([
                MyInt(1),
                MyList([
                    'foo',
                    MyTuple([MyFloat(99.9), MyList([1, 2, 3])]),
                    MyDict({123: 456}),
                ]),
                MyFloat(100.0),
            ]),
            (
                1,
                [
                    'foo',
                    (99.9, [1, 2, 3]),
                    {123: 456},
                ],
                100.0,
            ),
        ),
        (
            MyDict({
                MyStr('foo'): MyList([MyStr('a'), MyFloat(1.23), MyBytes(b'c')]),
                MyTuple(['f', 'o']): MyDict({MyInt(1): MyFloat(1.0)}),
                True: False,
            }),
            {
                'foo': ['a', 1.23, b'c'],
                ('f', 'o'): {1: 1.0},
                True: False,
            },
        ),
        (MyStr('foo'), 'foo'),
        (MyBytes(b'foo'), b'foo'),
        (MyInt(123), 123),
        (MyFloat(1.23), 1.23),
        (range(100), TypeError('Unsupported basic type: range: range(0, 100)')),
    ),
    ids=lambda v: repr(v),
)
def test_convert_to_basic_type(data, exp_result, mocker):
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            _utils.convert_to_basic_type(data)
    else:
        return_value = _utils.convert_to_basic_type(data)
        assert return_value == exp_result



import abc
import asyncio
import xmlrpc.client

from . import _base, _errors, _utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


class RtorrentURL(_utils.URL):
    """rTorrent RPC URL"""

    default = 'scgi://127.0.0.1:5000'

    @property
    def scheme(self):
        """Valid schemes: ``file``, ``scgi``, ``http``, ``https``"""
        return super().scheme

    @scheme.setter
    def scheme(self, scheme):
        if not scheme or str(scheme).lower() in ('file', 'scgi', 'http', 'https'):
            _utils.URL.scheme.fset(self, scheme)
        else:
            raise _errors.ValueError('Scheme must be "file", "scgi", "http" or "https"')

    @_utils.URL.host.setter
    def host(self, host):
        if host and self.scheme == 'file':
            raise _errors.ValueError(f"{self.scheme} URLs don't have a host")
        else:
            _utils.URL.host.fset(self, host)

    @_utils.URL.port.setter
    def port(self, port):
        if port and self.scheme == 'file':
            raise _errors.ValueError(f"{self.scheme} URLs don't have a port")
        else:
            _utils.URL.port.fset(self, port)

    @_utils.URL.username.setter
    def username(self, username):
        if username and self.scheme in ('file', 'scgi'):
            raise _errors.ValueError(f"{self.scheme} URLs don't have a username")
        else:
            _utils.URL.username.fset(self, username)

    @_utils.URL.password.setter
    def password(self, password):
        if password and self.scheme in ('file', 'scgi'):
            raise _errors.ValueError(f"{self.scheme} URLs don't have a password")
        else:
            _utils.URL.password.fset(self, password)

    @_utils.URL.path.setter
    def path(self, path):
        if path and self.scheme == 'scgi':
            raise _errors.ValueError(f"{self.scheme} URLs don't have a path")
        else:
            _utils.URL.path.fset(self, path)


class RtorrentRPC(_base.RPCBase):
    """
    RPC client for rTorrent

    URL formats:
        * ``[scgi://]HOST[:PORT]``
        * ``[file://]SOCKET_PATH``
        * ``http[s]://[USERNAME:PASSWORD@]HOST[:PORT][/PATH]``

    References:
        * https://github.com/rakshasa/rtorrent/wiki/rTorrent-0.9-Comprehensive-Command-list-(WIP)
        * https://docs.python.org/3/library/xmlrpc.client.html
        * https://github.com/rakshasa/rtorrent/wiki/RPC-Setup-XMLRPC

    :raise ValueError: if any argument is invalid
    """

    name = 'rtorrent'
    label = 'rTorrent'
    URL = RtorrentURL

    def __init__(
        self,
        url=None,
        *,
        scheme=None,
        host=None,
        port=None,
        username=None,
        password=None,
        timeout=None,
        proxy_url=None,
    ):
        # Set custom or default URL
        self.url = url

        # Update URL
        if scheme is not None:
            self.url.scheme = scheme
        if host is not None:
            self.url.host = host
        if port is not None:
            self.url.port = port
        if username is not None:
            self.url.username = username
        if password is not None:
            self.url.password = password

        self.timeout = timeout
        self.proxy_url = proxy_url

    async def _connect(self):
        # Close old proxy
        await self._disconnect()

        # Create new XMLRPC proxy
        self._xmlrpc = _AsyncServerProxy(
            url=self.url,
            proxy_url=self.proxy_url,
        )

        # Maybe raise connection/authentication error
        await self._call('system.pid')

    async def _disconnect(self):
        if hasattr(self, '_xmlrpc'):
            await self._xmlrpc.close()
            delattr(self, '_xmlrpc')

    async def _call(self, method, *args):
        try:
            return await _utils.catch_connection_exceptions(
                self._xmlrpc.call(method, *args),
            )

        except xmlrpc.client.ProtocolError as e:
            if e.errcode == 401:
                raise _errors.AuthenticationError('Authentication failed')
            else:
                msg = e.errmsg if e.errmsg else str(e)
                raise _errors.RPCError(msg)

        except xmlrpc.client.Fault as e:
            raise _errors.RPCError(e.faultString)

    async def multicall(self, *calls, raise_errors=True, as_dict=False):
        """
        Make ``system.multicall`` RPC request

        :param calls: ``(method, parameter1, parameter2, ...)`` tuples
        :param bool raise_errors: Whether to raise the first error response as
            :class:`~.RPCError` or return it like a normal response
        :param bool as_dict: Whether to map method names to return values

            .. note:: Every method can only be called once if this is enabled.

        :raise RPCError: if `raise_errors` is `True` and any response contains a
            dictionary with a ``faultString`` key

        :return: :class:`list` or :class:`dict` of return values
        """
        def pretty_call(i):
            method = calls[i][0]
            params = ', '.join(f'{param!r}' for param in calls[i][1:])
            return f'{method}({params})'

        if as_dict:
            # Because method names are dictionary keys, every method can only be
            # called once.
            methods = [call[0] for call in calls]
            indexes_map = {
                method: [i for i in range(len(methods))
                         if methods[i] == method]
                for method in methods
            }
            for method, indexes in indexes_map.items():
                if len(indexes) > 1:
                    calls = ', '.join(pretty_call(i) for i in indexes)
                    raise RuntimeError(f'Multiple {method} calls: {calls}')

        responses = await self.call('system.multicall', [
            {'methodName': method, 'params': params}
            for method, *params in calls
        ])

        return_values = []
        for i in range(len(responses)):
            response = responses[i]

            if isinstance(response, dict) and 'faultString' in response:
                exc = _errors.RPCError(str(response['faultString']))
                if raise_errors:
                    raise exc
                else:
                    return_values.append(exc)

            elif isinstance(response, list) and len(response) == 1:
                return_values.append(response[0])

            else:
                raise RuntimeError(f'Unexpected response: {response!r}')

        assert len(return_values) == len(calls)

        if as_dict:
            return {
                method: return_value
                for (method, *params), return_value in zip(calls, return_values)
            }
        else:
            return return_values

    _supported_methods = ()

    async def get_supported_method(self, *candidates):
        """
        Get first supported method from multiple candidates

        :param candidates: Sequence of method names

        :raise ValueError: if no method name in `candidates` is supported
        """
        if not self._supported_methods:
            type(self)._supported_methods = await self.call('system.listMethods')

        for method in candidates:
            if method in self._supported_methods:
                return method

        raise _errors.ValueError('Unsupported method(s): ' + ', '.join(f'{c!r}' for c in candidates))


class _AsyncServerProxy:
    def __init__(self, url, proxy_url=None):
        if url.scheme in ('http', 'https'):
            self._transport = _HttpTransport(
                url=url,
                proxy_url=proxy_url,
            )
        elif url.scheme == 'scgi':
            self._transport = _ScgiHostTransport(
                url=url,
                proxy_url=proxy_url,
            )
        elif url.scheme == 'file' and url.path:
            if proxy_url:
                raise _errors.ValueError(f'You cannot use a proxy to connect to {url}')
            else:
                self._transport = _ScgiSocketTransport(url=url)
        else:
            raise _errors.ValueError(f'Unsupported protocol: {url}')

    async def call(self, method_name, *params):
        request_data = xmlrpc.client.dumps(
            _utils.convert_to_basic_type(params),
            method_name,
            encoding='utf-8',
            allow_none=False,
        ).encode('utf-8', 'xmlcharrefreplace')

        # Return asynchronous iterator over chunks of bytes
        chunks = self._transport.request(request_data)
        return await self._parse_response(chunks)

    async def _parse_response(self, chunks):
        p, u = xmlrpc.client.getparser()
        async for chunk in chunks:
            p.feed(chunk)
        p.close()
        return_value = u.close()

        if len(return_value) == 1:
            return return_value[0]
        else:
            return return_value

    async def close(self):
        await self._transport.close()


class TransportBase(abc.ABC):
    @abc.abstractmethod
    async def request(self, data):
        """
        Send request and return asynchronous iterator over chunks of bytes
        """

    @abc.abstractmethod
    async def close(self):
        """Close any existing connections"""


class _HttpTransport(TransportBase):
    """Connect via reverse HTTP proxy"""

    def __init__(self, url, proxy_url=None):
        if url.scheme not in ('http', 'https'):
            raise _errors.ValueError(f'Unsupported protocol: {url.scheme}')
        else:
            if not url.path:
                url.path = '/RPC2'
            # Username and password are stored in self._http_client
            self._url = url.without_auth

        self._request_lock = asyncio.Lock()
        self._http_client = _utils.create_http_client(
            auth=(url.username, url.password),
            proxy_url=proxy_url.with_auth if proxy_url else None,
        )

    async def close(self):
        async with self._request_lock:
            await self._http_client.aclose()

    async def request(self, data):
        async with self._request_lock:
            aiterator = self._request(data)
            async for chunk in aiterator:
                yield chunk

    async def _request(self, data):
        async with self._http_client.stream('POST', self._url, content=data) as response:
            if response.status_code != 200:
                raise xmlrpc.client.ProtocolError(
                    url=self._url,
                    errcode=response.status_code,
                    errmsg=response.reason_phrase,
                    headers=response.headers,
                )
            else:
                aiterator = response.aiter_bytes()
                async for chunk in aiterator:
                    yield chunk


class _ScgiTransportBase(TransportBase, abc.ABC):
    """Base class for SCGI transports (network.scgi.*)"""

    async def close(self):
        pass

    async def request(self, data):
        reader, writer = await self._get_reader_writer()
        await self._send(writer, data)
        async for chunk in self._read(reader, writer, 1024):
            yield chunk

    @abc.abstractmethod
    async def _get_reader_writer(self):
        """Return `(:class:`StreamReader`, :class:`StreamWriter`)` tuple"""

    async def _read(self, reader, writer, chunk_size):
        try:
            headers_delim = b'\r\n\r\n'
            headers_done = False
            combined_headers = b''

            while True:
                chunk = await reader.read(chunk_size)
                if not chunk:
                    break
                elif headers_done:
                    # Headers are already fully read
                    yield chunk
                else:
                    combined_headers += chunk
                    if headers_delim in combined_headers:
                        # Find and remove HTTP headers
                        payload_start = combined_headers.index(headers_delim) + len(headers_delim)
                        first_payload_chunk = combined_headers[payload_start:]
                        if first_payload_chunk:
                            yield first_payload_chunk
                        headers_done = True
                        combined_headers = b''
        finally:
            writer.close()
            await writer.wait_closed()

    async def _send(self, writer, data):
        data_encoded = self._encode_request(data)
        writer.write(data_encoded)
        await writer.drain()

    def _encode_request(self, data):
        def encode_header(key, value):
            return key + b'\x00' + value + b'\x00'

        headers = (
            encode_header(b'CONTENT_LENGTH', str(len(data)).encode('utf-8'))
            + encode_header(b'SCGI', b'1')
            + encode_header(b'REQUEST_METHOD', b'POST')
            + encode_header(b'REQUEST_URI', self._path)
        )
        request = (
            str(len(headers)).encode('utf-8')
            + b':'
            + headers
            + b','
            + data
        )
        return request


class _ScgiHostTransport(_ScgiTransportBase):
    """Connect directly to rTorrent (network.scgi.open_port)"""

    def __init__(self, url, proxy_url=None):
        if url.scheme != 'scgi':
            raise _errors.ValueError(f'Unsupported protocol: {url.scheme}')
        self._host = url.host
        if not url.port:
            raise _errors.ValueError('No port specified')
        else:
            self._port = int(url.port)
        self._path = (url.path or '/RPC2').encode('utf-8')
        self._proxy_url = proxy_url

    async def _get_reader_writer(self):
        if self._proxy_url:
            import python_socks.async_.asyncio  # isort:skip
            try:
                proxy = python_socks.async_.asyncio.Proxy.from_url(self._proxy_url.with_auth)
            except ValueError as e:
                raise _errors.ValueError(e)
            sock = await proxy.connect(
                dest_host=self._host,
                dest_port=self._port,
                # Timeouts are handled with async_timeout in RPCBase
                timeout=float('inf'),
            )
            open_connection_kwargs = {
                'sock': sock,
            }
        else:
            open_connection_kwargs = {
                'host': self._host,
                'port': self._port,
            }

        reader, writer = await asyncio.open_connection(**open_connection_kwargs)
        return reader, writer


class _ScgiSocketTransport(_ScgiTransportBase):
    """Connect directly to rTorrent (network.scgi.open_local)"""

    def __init__(self, url):
        if url.scheme != 'file':
            raise _errors.ValueError(f'Unsupported protocol: {url.scheme}')
        self._socket_path = url.path
        self._path = b'/RPC2'

    async def _get_reader_writer(self):
        reader, writer = await asyncio.open_unix_connection(path=self._socket_path)
        return reader, writer
